/*
 *  DNSpoison
 *  src/netconf.c
 *
 *  Copyright (C) 2011
 *
 *  Author: Vilmain Nicolas
 *  Contact: null.sim@gmail.com
 *
 *  This file is part of DNSpoison.
 *
 *  DNSpoison is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DNSpoison is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DNSpoison.  If not, see <http://www.gnu.org/licenses/>.
 */

#include  "dnspoison.h"

static int   get_interface (struct netconf_s *nc, struct ifaddrs *ifa,
			    const char *name);
static int   get_interface_address (struct ifaddrs *ifa, struct netconf_s *nc);
static int   get_interface_more_data (struct netconf_s *nc);

int
get_netconf (struct netconf_s *nc, const char *name)
{
  int ret;
  struct ifaddrs *ifa = NULL;

  memset (nc, 0, sizeof (struct netconf_s));
  if (getifaddrs (&ifa) == -1)
    {
      fprintf (stderr, "%s:getifaddrs: %s\n",
	       program_name, strerror (errno));
      return -1;
    }
  ret = get_interface (nc, ifa, name);
  freeifaddrs (ifa);
  return ret;
}

static int
get_interface (struct netconf_s *nc, struct ifaddrs *ifa,
	       const char *name)
{
  struct ifaddrs  *p_ifa = NULL;

  p_ifa = ifa;
  while (p_ifa)
    {
      if (!(p_ifa->ifa_flags & IFF_LOOPBACK)
	  && (p_ifa->ifa_flags & IFF_RUNNING)
	  && (p_ifa->ifa_addr)
	  && (p_ifa->ifa_addr->sa_family == AF_INET)
	  && ((!name || !name[0])
	      || !strcmp (name, p_ifa->ifa_name)))
	{
	  strncpy (nc->nc_name, p_ifa->ifa_name, (NAME_LEN - 1));
	  return (get_interface_address (ifa, nc) == -1
		  || get_interface_more_data (nc) == -1) ? -1 : 0;
	}
      p_ifa = p_ifa->ifa_next;
    }
  fprintf (stderr, "%s: cannot find or valid network interface\n",
	   program_name);
  return -1;
}

static int
get_interface_address (struct ifaddrs *ifa, struct netconf_s *nc)
{
  while (ifa)
    {
      if (!strcmp (ifa->ifa_name, nc->nc_name))
	{
	  if (ifa->ifa_addr->sa_family == AF_INET)
	    {
	      memcpy (&nc->nc_ipv4,
		      &((struct sockaddr_in *) ifa->ifa_addr)->sin_addr,
		      sizeof (struct in_addr));
	    }
	}
      ifa = ifa->ifa_next;
    }
  return 0;
}

static int
get_interface_more_data (struct netconf_s *nc)
{
  int           fd;
  struct ifreq  ifr;

  fd = socket (AF_INET, SOCK_DGRAM, 0);
  if (fd == -1)
    {
      fprintf (stderr, "%s:socket: %s\n", program_name,
	       strerror (errno));
      return -1;
    }
  memset (&ifr, 0, sizeof (struct ifreq));
  strncpy (ifr.ifr_name, nc->nc_name, (NAME_LEN - 1));
  if (ioctl (fd, SIOCGIFMTU, &ifr) == -1)
    {
      fprintf (stderr, "%s:ioctl(SIOCGIFMTU): %s\n",
	       program_name, strerror (errno));
      close (fd);
      return -1;
    }
  nc->nc_mtu = ifr.ifr_ifru.ifru_mtu;
  if (ioctl (fd, SIOCGIFINDEX, &ifr) == -1)
    {
      fprintf (stderr, "%s:ioctl(SIOCGIFINDEX): %s\n",
	       program_name, strerror (errno));
      close (fd);
      return -1;
    }
#ifdef __BSD
  nc->nc_index = ifr.ifr_index;
#elif defined(__Linux)
  nc->nc_index = ifr.ifr_ifindex;
#endif /* __Linux */
  close (fd);
  return 0;
}
