/*
 *  DNSpoison
 *  src/socket.c
 *
 *  Copyright (C) 2011
 *
 *  Author: Vilmain Nicolas
 *  Contact: null.sim@gmail.com
 *
 *  This file is part of DNSpoison.
 *
 *  DNSpoison is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DNSpoison is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DNSpoison.  If not, see <http://www.gnu.org/licenses/>.
 */

#include  "dnspoison.h"

static int   recv_timeout (int fd);
#ifdef __BSD
static int    get_bpf (const char *name);
static int    conf_bpf (int fd, const char *device_name);
#endif /* __BSD */
static int    xread (int fd, void *buf, size_t size);

int
get_ethsock (struct ethsock_s *s, int index, const char *name)
{

#ifdef __Linux
  (void) name;
  s->fdsock = socket (AF_PACKET, SOCK_RAW, ETH_P_ALL);
  if (s->fdsock == -1)
    {
      fprintf (stderr, "%s:socket: %s\n", program_name,
	       strerror (errno));
      return -1;
    }
  memset (&s->ll, 0, SOCK_LL_SIZE);
  s->ll.sll_family = AF_PACKET;
  s->ll.sll_protocol = htons (ETH_P_ALL);
  s->ll.sll_ifindex = index;
  if (bind (s->fdsock, (struct sockaddr *) &s->ll, SOCK_LL_SIZE) == -1)
    {
      fprintf (stderr, "%s:bind: %s\n", program_name, strerror (errno));
      close (s->fdsock);
      return -1;
    }

#elif defined(__BSD)
  (void) index;
  s->fdsock = get_bpf (name);
  if (s->fdsock == -1)
    return -1;
#endif /* not __Linux */

  return 0;
}

int
ethsend (struct ethsock_s *s, void *packet, size_t size)
{
  int n_send;

#ifdef __Linux
  n_send = sendto (s->fdsock, packet, size, 0,
		   (struct sockaddr *) &s->ll, SOCK_LL_SIZE);
  if (n_send == -1)
    fprintf (stderr, "%s:send: %s\n", program_name, strerror (errno));
#elif defined (__BSD)
  n_send = write (s->fdsock, packet, size);
  if (n_send == -1)
    fprintf (stderr, "%s:write: %s\n", program_name, strerror (errno));
#endif /* not __Linux */
  else if ((size_t) n_send != size)
    fprintf (stderr, "%s:WARNING: maybe packet send is troncated\n",
	     program_name);
  return n_send;
}

int
ethrecv (int fdsock, union packet *pkt, int max_size)
{
  int ret;
  static char st_packet[PACKET_SIZE];
#ifdef __BSD
  static size_t offset = 0;
  static size_t totsize = 0;
  struct bpf_hdr *bpf = NULL;
#endif /* __BSD */

  for (;;)
    {
#ifdef __Linux
      ret = recv_timeout (fdsock);
      if (ret == -1)
	return -1;
      ret = xread (fdsock, st_packet, PACKET_SIZE);
      if (ret == -1)
	return -1;
      if (ret > (int) MIN_PACKET && ret < max_size)
	{
	  memcpy (pkt->packet , st_packet, ret);
#elif defined (__BSD)
      if (offset >= totsize)
	{
	  offset = 0;
	  totsize = 0;
	  totsize = recv_timeout (fdsock);
	  if ((int) totsize == -1)
	    return -1;
	  memset (st_packet, 0, PACKET_SIZE);
	  totsize = xread (fdsock, st_packet, PACKET_SIZE);
	  if ((int) totsize == -1)
	    return -1;
	}
      bpf = (struct bpf_hdr *) (st_packet + offset);
      if (bpf->bh_caplen > (int) MIN_PACKET && (int) bpf->bh_caplen < max_size)
	{
	  ret = bpf->bh_caplen;
	  memcpy (pkt->packet, (st_packet + offset
				+ bpf->bh_hdrlen), bpf->bh_caplen);
#endif /* not __Linux */
	  if (pkt->hdr.eth.protocol == htons (ETH_IP)
	      && pkt->hdr.ip.protocol == PROTO_UDP
	      && pkt->hdr.udp.dest == htons (53))
	    {
#ifdef __BSD
	      offset += BPF_WORDALIGN(bpf->bh_hdrlen + bpf->bh_caplen);
#endif /* __BSD */
	      return ret;
	    }
	}
#ifdef __BSD
      offset += BPF_WORDALIGN(bpf->bh_hdrlen + bpf->bh_caplen);
#endif /* __BSD */
    }
  return 0;
}

static int
recv_timeout (int fd)
{
  fd_set read_ok; 

  FD_ZERO (&read_ok);
  FD_SET (fd, &read_ok);
  if (select (fd + 1, &read_ok, NULL, NULL, NULL) == -1)
    {
      if (errno != EINTR)
	fprintf (stderr, "%s:select: %s\n", program_name,
		 strerror (errno));
      return -1;
    }
  return 0;
}

#ifdef __BSD
static int
get_bpf (const char *name)
{
  int   n;
  int   bpf_fd;
  char  bpf_path[15];

  n = 0;
  do
    {
      snprintf (bpf_path, sizeof (bpf_path) - 1,
		"/dev/bpf%d", n);
      bpf_fd = open (bpf_path, O_RDWR);
    }
  while (bpf_fd == -1 && ++n < 10);
  if (bpf_fd == -1)
    {
      fprintf (stderr, "cannot find bpf device\n");
      return -1;
    }
  if (conf_bpf (bpf_fd, name))
    {
      close (bpf_fd);
      return -1;
    }
  return bpf_fd;
}

static int
conf_bpf (int fd, const char *device_name)
{
  int bpf_opt;
  struct ifreq  ifr;

  bpf_opt = PACKET_SIZE;
  if (ioctl (fd, BIOCSBLEN, &bpf_opt) == -1)
    {
      fprintf (stderr, "%s:ioctl(BIOCSBLEN): %s\n",
	       program_name, strerror (errno));
      return -1;
    }
  strncpy (ifr.ifr_name, device_name, 10);
  if (ioctl (fd, BIOCSETIF, &ifr) == -1)
    {
      fprintf (stderr, "%s:ioctl(BIOCSETIF): %s\n",
	       program_name, strerror (errno));
      return -1;
    }
  bpf_opt = 1;
  if (ioctl (fd, BIOCIMMEDIATE, &bpf_opt) == -1)
    {
      fprintf (stderr, "%s:ioctl(BIOCIMMEDIATE): %s\n",
	       program_name, strerror (errno));
      return -1;
    }
  return 0;
}
#endif /* __BSD */

static int
xread (int fd, void *buf, size_t size)
{
  int ret;

  do
    ret = read (fd, buf, size);
  while (ret == -1 && errno == EINTR);
  if (ret == -1)
    fprintf (stderr, "%s:read: %s\n", program_name, strerror (errno));
  return ret;
}
